/*******************************************************
 * Lexer for VRML++ Preprocessor
 * Copyright (C) 1996 Stephan Diehl
 * 
 * extends the VRML 2.0 Lexer of Bell and Woods,
 * (see Copyright message below)
 * 
 * extensions are enclosed in #ifdef STED ... #endif
 *       or comments << STED  ...  STED >> 
 *******************************************************/


/**************************************************
 * VRML 2.0 Parser
 * Copyright (C) 1996 Silicon Graphics, Inc.
 *
 * Author(s)	: Gavin Bell
 *                Daniel Woods (first port)
 **************************************************
 */
%{

#include "parstree.h"
#include "tokens.h"
#include <string.h>

#ifdef __cplusplus
extern "C" int yylex  ();
#else
extern int yylex ( );
#endif


	/* Current line number */
int currentLineNumber = 1;
int column=0;
void yyResetLineNumber() { currentLineNumber = 1; column=0; }

#ifdef __cplusplus
extern "C" void yyerror (const char *);
#else
extern void yyerror (const char *);
#endif

/* extern void yyerror(const char *);
*/

    /* The YACC parser sets this to a token to direct the lexer */
    /* in cases where just syntax isn't enough: */
int expectToken = 0;

    /* True when parsing a multiple-valued field: */
static int parsing_mf = 0;

   /* These are used when parsing SFImage fields: */
static int sfImageIntsParsed = 0;
static int sfImageIntsExpected = 0;

#ifdef STED
#define YY_SKIP_YYWRAP
#endif

int yywrap() { BEGIN INITIAL; return 1; } 

extern char* linebuffer;

int count()
  { char* s=yytext;
    linebuffer[column]=*s;
    for(;*s;s++,linebuffer[column]=*s) 
     { if ((*s)=='\n') 
        { currentLineNumber++; column=0; }
       else if ((*s)=='\t')
            column += 8 - (column % 8);
       else column++;
     }
  }
  
int newlines_in(char* s)
 { int res=0;
   for(;*s;s++) if ((*s)=='\n') res++;
   return res; 
 }

%}

	/* Normal state:  parsing nodes.  The initial start state is used */
    /* only to recognize the VRML header. */
%x NODE

	/* Start tokens for all of the field types, */
    /* except for MFNode and SFNode, which are almost completely handled */
    /* by the parser: */
%x SFB SFC SFF SFIMG SFI SFR SFS SFT SFV2 SFV3
%x MFC MFF MFI MFR MFS MFV2 MFV3
%x IN_SFS IN_MFS IN_SFIMG

	
/* Big hairy expression for floating point numbers: */

/* STED: had to change this one, didn't work for e.g. -2e3
  float (-?(([0-9]+)|([0-9]*\.[0-9]+)([eE][+\-]?[0-9]+)?))
*/
float (-?(([0-9]+)|([0-9]*\.[0-9]+))([eE][+\-]?[0-9]+)?)

	/* Ints are decimal or hex (0x##): */
int (-?([0-9]+)|(0[xX][0-9a-fA-F]*))

	/* Whitespace.  Using this pattern can screw up currentLineNumber, */
    /* so it is only used wherever it is really convenient and it is */
    /* extremely unlikely that the user will put in a carriage return */
    /* (example: between the floats in an SFVec3f) */
ws ([ \t\r\n,]|(#.*))+
	/* And the same pattern without the newline */
wsnnl ([ \t\r,]|(#.*))

	/* Legal characters to start an identifier */
idStartChar ([^\x30-\x39\x00-\x20\x22\x23\x27\x2b-\x2e\x5b-\x5d\x7b\x7d])
	/* Legal other characters in an identifier */
idRestChar	([^\x00-\x20\x22\x23\x27\x2b-\x2e\x5b-\x5d\x7b\x7d])
%%

%{
	/* Switch into a new start state if the parser */
    /* just told us that we've read a field name */
    /* and should expect a field value (or IS) */
     if (expectToken != 0) {
      if (yy_flex_debug)
          fprintf(stderr,"LEX--> Start State %d\n", expectToken);
      
      /*
       * Annoying.  This big switch is necessary because
       * LEX wants to assign particular numbers to start
       * tokens, and YACC wants to define all the tokens
       * used, too.  Sigh.
       */
      switch(expectToken) {
       	case SFBOOL: BEGIN SFB; break;
        case SFCOLOR: BEGIN SFC; break;
        case SFFLOAT: BEGIN SFF; break;
        case SFIMAGE: BEGIN SFIMG; break;
        case SFINT32: BEGIN SFI; break;
        case SFROTATION: BEGIN SFR; break;
        case SFSTRING: BEGIN SFS; break;
        case SFTIME: BEGIN SFT; break;
        case SFVEC2F: BEGIN SFV2; break;
        case SFVEC3F: BEGIN SFV3; break;
        case MFCOLOR: BEGIN MFC; break;
        case MFFLOAT: BEGIN MFF; break;
        case MFINT32: BEGIN MFI; break;
        case MFROTATION: BEGIN MFR; break;
        case MFSTRING: BEGIN MFS; break;
        case MFVEC2F: BEGIN MFV2; break;
        case MFVEC3F: BEGIN MFV3; break;

        /* SFNode and MFNode are special.  Here the lexer just returns */
        /* "marker tokens" so the parser knows what type of field is */
        /* being parsed; unlike the other fields, parsing of SFNode/MFNode */
        /* field happens in the parser. */
        case MFNODE: expectToken = 0; return MFNODE;
        case SFNODE: expectToken = 0; return SFNODE;
        
        default: yyerror("ACK: Bad expectToken"); break;
      }
    }
%}

	/* This is more complicated than they really need to be because */
    /* I was ambitious and made the whitespace-matching rule aggressive */

   /* STED new INITIAL state for VRML++*/

<INITIAL>"#VRML V2.0 utf8".*\n{wsnnl}*	{ BEGIN NODE; count(); }
<INITIAL>"#VRML++ draft utf8".*\n{wsnnl}*  { BEGIN NODE; count(); }



	/* The lexer is in the NODE state when parsing nodes, either */
    /* top-level nodes in the .wrl file, in a prototype implementation, */
    /* or when parsing the contents of SFNode or MFNode fields. */
<NODE>PROTO		{ count(); return PROTO; }
<NODE>EXTERNPROTO	{ count(); return EXTERNPROTO; }

<NODE>CLASS		{ count(); return CLASS; }
<NODE>EXTERNCLASS	{ count(); return EXTERNCLASS; }
<NODE>EXTENDS		{ count(); return EXTENDS; }

<NODE>DEF	        { count(); return DEF; }
<NODE>USE		{ count(); return USE; }
<NODE>TO		{ count(); return TO; }
<NODE>IS		{ count(); return IS; }
<NODE>ROUTE		{ count(); return ROUTE; }
<NODE>UNROUTE		{ count(); return UNROUTE; }
<NODE>NULL		{ count(); return SFN_NULL; }
<NODE>eventIn		{ count(); return EVENTIN; }
<NODE>eventOut		{ count(); return EVENTOUT; }
<NODE>field		{ count(); return FIELD; }
<NODE>exposedField	{ count(); return EXPOSEDFIELD; } 

	/* Legal identifiers: */
<NODE>{idStartChar}{idRestChar}*	{ count();
                                          yylval.string = strdup(yytext);
			       		  return IDENTIFIER; }

	/* All fields may have an IS declaration: */
<SFB,SFC,SFF,SFIMG,SFI,SFR,SFS,SFT,SFV2,SFV3>IS	   { BEGIN NODE;
                                                     expectToken = 0;
                                                     yyless(0);
                                                   }
<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>IS     { BEGIN NODE;
		                        expectToken = 0;
                                        yyless(0); /* put back the IS */
                                      }

    /* All MF field types other than MFNode are completely parsed here */
    /* in the lexer, and one token is returned to the parser.  They all */
    /* share the same rules for open and closing brackets: */
<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>\[    { if (parsing_mf) yyerror("Double [");
				       parsing_mf = 1;
                                       yymore();
                                     }
<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>\]    { if (!parsing_mf) yyerror("Unmatched ]");
				       int fieldType = expectToken;
				       BEGIN NODE;
                                       parsing_mf = 0;
                                       expectToken = 0;
  /* STED */
                               yylval.pt=new ParseTreeNode(yytext,fieldType);
                               count();
                               return fieldType;
                            }
                                      

<SFB>TRUE		{ BEGIN NODE; expectToken = 0; 
                          yylval.pt=new ParseTreeNode(yytext,SFBOOL);
                          count();
                          return SFBOOL; }
<SFB>FALSE		{ BEGIN NODE; expectToken = 0;
                          yylval.pt=new ParseTreeNode(yytext,SFBOOL);
                          count();
                          return SFBOOL; }

<SFI>{int}		{ BEGIN NODE; expectToken = 0; 
                          yylval.pt=new ParseTreeNode(yytext,SFINT32);
                          count();
                          return SFINT32; }
<MFI>{int}		{ if (parsing_mf) { yymore();}
	            /* A real implementation would add the float to an */
                    /* array...  but here we do nothing */
				  else {
                  	  BEGIN NODE; expectToken = 0;
                          yylval.pt=new ParseTreeNode(yytext,MFINT32);
                          count();
                          return MFINT32; 
                  }
                }

	/* All the floating-point types are pretty similar: */
<SFF>{float}		{ BEGIN NODE; expectToken = 0;
                          yylval.pt=new ParseTreeNode(yytext,SFFLOAT);
                          count();
                          return SFFLOAT; }
<MFF>{float}		{ if (parsing_mf) {yymore();} /* Add to array... */
                  	  else {
                      	/* No open bracket means a single value: */
                        BEGIN NODE; expectToken = 0; 
                        yylval.pt=new ParseTreeNode(yytext,MFFLOAT);
                        count();
                        return MFFLOAT;
                      }
                    }
<SFV2>{float}{ws}{float}    { BEGIN NODE; expectToken = 0;
                              yylval.pt=new ParseTreeNode(yytext,SFVEC2F);
                              count();
                              return SFVEC2F; }
<MFV2>{float}{ws}{float}   {  if (parsing_mf) {yymore();} /* add to array... */
              		      else {
                              	BEGIN NODE; expectToken = 0; 
                                yylval.pt=new ParseTreeNode(yytext,MFVEC2F);
                                count();
                                return MFVEC2F;
                              }
                            }
<SFV3>({float}{ws}){2}{float}	{ BEGIN NODE; expectToken = 0; 
                                  yylval.pt=new ParseTreeNode(yytext,SFVEC3F);
                                  count();
                                  return SFVEC3F; }
<MFV3>({float}{ws}){2}{float}	{ if (parsing_mf) {yymore();} /* add to array... */
              			  else {
                                  BEGIN NODE; expectToken = 0; 
                                  yylval.pt=new ParseTreeNode(yytext,MFVEC3F);
                                  count();
                                  return MFVEC3F;
                                  }
                                }
<SFR>({float}{ws}){3}{float}    { BEGIN NODE; expectToken = 0; 
                                  yylval.pt=new ParseTreeNode(yytext,SFROTATION);
                                  count();
                                  return SFROTATION; }
<MFR>({float}{ws}){3}{float}	{ if (parsing_mf) {yymore();} /* .. add to array... */
				  else {
                                    BEGIN NODE; expectToken = 0; 
                                    yylval.pt=new ParseTreeNode(yytext,MFROTATION);
                                    count();
                                    return MFROTATION;
                                  }
                                }
<SFC>({float}{ws}){2}{float}    { BEGIN NODE; expectToken = 0; 
                                  yylval.pt=new ParseTreeNode(yytext,SFCOLOR);
                                  count();
                                  return SFCOLOR; }
<MFC>({float}{ws}){2}{float}	{ if (parsing_mf) {yymore();} /* .. add to array... */
			          else {
                                    BEGIN NODE; expectToken = 0; 
                                    yylval.pt=new ParseTreeNode(yytext,MFCOLOR);
                                    count();
                                    return MFCOLOR;
                                  }
                                }
<SFT>{float}	    { BEGIN NODE; expectToken = 0;
                      yylval.pt=new ParseTreeNode(yytext,SFTIME);
                      count();
                      return SFTIME; }
               
    /* SFString/MFString */
<SFS>\"				{    yymore(); BEGIN IN_SFS; }
<MFS>\"				{    yymore(); BEGIN IN_MFS; }
	/* Anything besides open-quote (or whitespace) is an error: */
<SFS>[^ \"\t\r\,\n]+	{ yyerror("String missing open-quote");
			  BEGIN NODE; expectToken = 0; 
                          count();
                          return SFSTRING;
                        }
    /* Expect open-quote, open-bracket, or whitespace: */
<MFS>[^ \[\]\"\t\r\,\n]+	{ yyerror("String missing open-quote");
			  	  BEGIN NODE; expectToken = 0; 
                                  count();
                                  return MFSTRING;
                            }

<MFC,MFF,MFI,MFR,MFS,MFV2,MFV3>[ \t\r\,\n]+  	{ yymore(); } 

    /* Backslashed-quotes are OK: */
<IN_SFS,IN_MFS>\\\"	  {    }   
    /* Gobble up anything besides quotes and newlines. */
    /* Newlines are legal in strings, but we exclude them here so */
    /* that line number are counted correctly by the catch-all newline */
    /* rule that applies to everything. */
 /* STED was <IN_SFS,IN_MFS>[^ \"\n]+ ; */
 /*   we need newlines in strings for source code */  
 /*   in Script nodes' url field	  */
 /* 
 /* <IN_SFS,IN_MFS>[^ \"]+      ; */ 

<IN_SFS,IN_MFS>[^ \"\n]+        { yymore(); } 

	/* Quote ends the string: */
<IN_SFS>\"		{ BEGIN NODE; expectToken = 0; 
                          yylval.pt=new ParseTreeNode(yytext,SFSTRING);
                          count();
                          return SFSTRING; }
<IN_MFS>\"              { if (parsing_mf) 
                              { yymore();   /* sted */
                                BEGIN MFS;
                                
                                /* ... add to array ... */
                              }
			    else  {
                          	  BEGIN NODE; expectToken = 0; 
                                  yylval.pt=new ParseTreeNode(yytext,MFSTRING);
                                  count();
                                  return MFSTRING;
                            }
                        }

<IN_SFS,IN_MFS>{wsnnl}*\n{wsnnl}*	{ yymore(); }
<IN_SFS,IN_MFS>.                	{ yymore(); }

    /* SFImage: width height numComponents then width*height integers: */
<SFIMG>{int}{ws}{int}	{ int w, h;
	       		  sscanf(yytext, "%d %d", &w, &h);
                          sfImageIntsExpected = 1+w*h;
                          sfImageIntsParsed = 0;   
                          BEGIN IN_SFIMG;
                        }
<IN_SFIMG>{int}         { ++sfImageIntsParsed;
                          if (sfImageIntsParsed == sfImageIntsExpected) 
                            {  	  BEGIN NODE; 
                                  expectToken = 0; 
                                  yylval.pt=new ParseTreeNode(yytext,SFIMAGE);
                                  count();
                                  return SFIMAGE;
                          }
                        }



	/* This is also whitespace, but we'll keep track of line number */
    /* to report in errors: */
<*>{wsnnl}*\n{wsnnl}*		{ count(); }
	/* Whitespace and catch-all rules apply to all start states: */
<*>{wsnnl}+		{ count(); }
	/* This catch-all rule catches anything not covered by any of */
    /* the above: */
<*>. 			{ count(); return yytext[0]; }




